

#include <stdlib.h>


typedef struct __attribute__((__packed__)) {
    float* input; size_t bytes_input;
    float* result; size_t bytes_result;
} 
RootIn;




void scale_values(float* input, size_t bytes_input) {

  __visc__hint(DEVICE);
  __visc__attributes(1, input, 1, input);
  
  for (int ind = 0; ind < bytes_input; ind++){
    input[ind] = input[ind] * 2.0;
  }
  
  __visc__return(1, bytes_input);
}




void graphRoot(/*0*/ float* input, /*1*/ size_t bytes_input, 
               /*2*/ float* result, /*3*/ size_t bytes_result) {

  //Specifies compilation target for current node
  __visc__hint(CPU_TARGET);

  __visc__attributes(2, input, result, 2, input, result);

  // Create an 0D (specified by 1st argument) HPVM node - so a single node
  // associated with node function ---_fxp_wrapper

  void* scaleNode = __visc__createNodeND(0, scale_values);
    
  // BindIn binds inputs of current node with specified node
  // - destination node
  // - argument position in argument list of function of source node
  // - argument position in argument list of function of destination node
  // - streaming (1) or non-streaming (0)

  // Edge transfers data between nodes within the same level of hierarchy.
  // - source and destination dataflow nodes
  // - edge type, all-all (1) or one-one(0)
  // - source position (in output struct of source node)
  // - destination position (in argument list of destination node)
  // - streaming (1) or non-streaming (0)

  // scale_fxp inputs
  __visc__bindIn(scaleNode, 0, 0, 0); // input -> ScNode:input
  __visc__bindIn(scaleNode, 1, 1, 0); // bytes_input -> ScNode:bytes_input
    
  // Similar to bindIn, but for the output. Output of a node is a struct, and
  // we consider the fields in increasing ordering.
  __visc__bindOut(scaleNode, 0, 0, 0);
    
}





int main(int argc, char* argv[]) {

    __visc__init();

    size_t input_size = 100;
    size_t result_size = 100;

    size_t input_bytes = input_size * sizeof(float);
    size_t result_bytes = result_size * sizeof(float);

    // This is host_input in cam_pipe()
    input = (float*) malloc(input_bytes);
    // This is host_result in cam_pipe()
    result = (float*) malloc(result_bytes);


    RootIn* rootArgs = (RootIn*) malloc(sizeof(RootIn));

    // Set up HPVM DFG inputs in the rootArgs struct.
    rootArgs->input = input;
    rootArgs->bytes_input = input_bytes;
    
    rootArgs->result = result;
    rootArgs->bytes_result = result_bytes;

    llvm_visc_track_mem(input, input_bytes);
    llvm_visc_track_mem(result, result_bytes);


    void* testDFG = __visc__launch(0, graphRoot, (void*) rootArgs);
    __visc__wait(testDFG);


    llvm_visc_request_mem(result, input_bytes);
    llvm_visc_request_mem(result, result_bytes);



    __visc__cleanup();

    return 0;
}



    
