#include <iostream>
#include <cstdio>
#include <cstring>
#include <cinttypes>
#include <visc.h>
#include <tensorTypes.h>
#include <tensorUtils.h>

using namespace std;

void tensorMulNode(void *t1, size_t bytes1, void *t2, size_t bytes2) {
    __visc__hint(visc::CUDNN_TARGET);
    __visc__attributes(2, t1, t2, 0);

    // X * W = t2 * t1
    void *r = __visc__tensor_mul(t2, t1);
    __visc__return(2, r, (size_t) 0);
}

void tensorAddNode(void *t1, size_t bytest1, void *t2, size_t bytest2) {
    __visc__hint(visc::CUDNN_TARGET);
    __visc__attributes(2, t1, t2, 0);

    void* r = __visc__tensor_add(t1, t2);
    __visc__return(2, r, (size_t) 0);
}

void root(void *w, size_t bytesw, void *x, size_t bytesx, void *b, size_t bytesb) {
    __visc__hint(visc::CPU_TARGET);
    __visc__attributes(3, w, x, b, 0);

    void *nodeMul = __visc__createNodeND(0, tensorMulNode);
    void *nodeAdd = __visc__createNodeND(0, tensorAddNode);

    // node, src, dst, stream
    __visc__bindIn(nodeMul, 0, 0, 0);
    __visc__bindIn(nodeMul, 1, 1, 0);
    __visc__bindIn(nodeMul, 2, 2, 0);
    __visc__bindIn(nodeMul, 3, 3, 0);

    // node, node, type, src, dst, stream
    __visc__edge(nodeMul, nodeAdd, 1, 0, 0, 0);
    __visc__edge(nodeMul, nodeAdd, 1, 1, 1, 0);

    __visc__bindIn(nodeAdd, 4, 2, 0);
    __visc__bindIn(nodeAdd, 5, 3, 0);

    __visc__bindOut(nodeAdd, 0, 0, 0);
    __visc__bindOut(nodeAdd, 1, 1, 0);

}


// Return type for the nodes
struct ret_t {
    void *tensor;
    size_t bytes;
};

typedef struct __attribute__((__packed__)) {
    void *w;
    size_t bytesw;
    void *x;
    size_t bytesx;
    void *b;
    size_t bytesb;
    struct ret_t r;
}
RootIn;

int main() {

    void *w;
    void *x;
    void *b;
    int test_batch_size = 1000;
    std::string prefix = "../../../../../../projects/hpvm-tensor-rt/model_params";
    std::string input_data_path = prefix + std::string("/FC_network2/mnist_float_input.bin");
    std::string W_path = prefix + std::string("/fc2_clipped/fc1.bin");			  
    std::string B_path = prefix + std::string("/fc2_clipped/fc1_bias.bin");  
    
    printf("Reading Input Data from = %s \n", input_data_path.c_str());
    
    x = readTrainedWeights(input_data_path.c_str(), float_type,
                           test_batch_size, 1, 28, 28);
    w = readTrainedWeights(W_path.c_str(), float_type, 1, 1, 784, 128);
    b = readTrainedWeights(B_path.c_str(), float_type, 1, 128, 1, 1);

    __visc__init();

    RootIn* args = static_cast<RootIn*>(malloc(sizeof(RootIn)));
    args->w = w;
    args->bytesw = 0;
    args->x = x;
    args->bytesx = 0;
    args->b = b;
    args->bytesb = 0;

    void *dfg = __visc__launch(0, root, (void *)args);

    __visc__wait(dfg);

    // FIXME: Value returned in the wrong index!!
    //void *r = static_cast<RootIn*>(args)->r.tensor;
    void *r = static_cast<RootIn*>(args)->w;
    hpvm_request_tensor(r, 0);

    __visc__cleanup();
    return 0;
}


