
#include <stdio.h> 
#include <stdlib.h> 
#include <unistd.h> 
#include <fcntl.h> 
#include <sys/types.h> 
#include <sys/stat.h> 
#include <string.h> 
#include "../../tensor_runtime/include/tensor_runtime.h" 
#include "../include/utils.h" 

int main(){ 

  llvm_hpvm_initTensorRt(0); 


  std::string dir_prefix = std::string("hpvm_mio_4/"); 
  std::string input_path =  dir_prefix + std::string("input.bin"); 
  std::string labels_path =  dir_prefix + std::string("labels.bin"); 
  std::string conv2d_1_w_path =  dir_prefix + std::string("conv2d_1_w.bin"); 
  void* conv2d_1_w =  readTrainedWeights(conv2d_1_w_path.c_str(), 0,32,3,3,3); 
  std::string conv2d_1_b_path =  dir_prefix + std::string("conv2d_1_b.bin"); 
  void* conv2d_1_b =  readTrainedWeights(conv2d_1_b_path.c_str(), 0,1,32,1,1); 
  std::string conv2d_2_w_path =  dir_prefix + std::string("conv2d_2_w.bin"); 
  void* conv2d_2_w =  readTrainedWeights(conv2d_2_w_path.c_str(), 0,32,32,3,3); 
  std::string conv2d_2_b_path =  dir_prefix + std::string("conv2d_2_b.bin"); 
  void* conv2d_2_b =  readTrainedWeights(conv2d_2_b_path.c_str(), 0,1,32,1,1); 
  std::string conv2d_3_w_path =  dir_prefix + std::string("conv2d_3_w.bin"); 
  void* conv2d_3_w =  readTrainedWeights(conv2d_3_w_path.c_str(), 0,64,32,3,3); 
  std::string conv2d_3_b_path =  dir_prefix + std::string("conv2d_3_b.bin"); 
  void* conv2d_3_b =  readTrainedWeights(conv2d_3_b_path.c_str(), 0,1,64,1,1); 
  std::string conv2d_4_w_path =  dir_prefix + std::string("conv2d_4_w.bin"); 
  void* conv2d_4_w =  readTrainedWeights(conv2d_4_w_path.c_str(), 0,64,64,3,3); 
  std::string conv2d_4_b_path =  dir_prefix + std::string("conv2d_4_b.bin"); 
  void* conv2d_4_b =  readTrainedWeights(conv2d_4_b_path.c_str(), 0,1,64,1,1); 
  std::string dense_1_w_path =  dir_prefix + std::string("dense_1_w.bin"); 
  void* dense_1_w =  readTrainedWeights(dense_1_w_path.c_str(), 0,1,1,1600,256); 
  std::string dense_1_b_path =  dir_prefix + std::string("dense_1_b.bin"); 
  void* dense_1_b =  readTrainedWeights(dense_1_b_path.c_str(), 0,1,256,1,1); 
  std::string dense_2_w_path =  dir_prefix + std::string("dense_2_w.bin"); 
  void* dense_2_w =  readTrainedWeights(dense_2_w_path.c_str(), 0,1,1,256,5); 
  std::string dense_2_b_path =  dir_prefix + std::string("dense_2_b.bin"); 
  void* dense_2_b =  readTrainedWeights(dense_2_b_path.c_str(), 0,1,5,1,1); 



  startMemTracking(); 

  int test_input_size = 5000; 
  int batch_size = 5000; 
  int batch_count = test_input_size / batch_size; 
  float final_accuracy = 0.0; 

  for(int i = 0; i < batch_count; i++){ 

    int start = i * batch_size; 
    int end = (i + 1) * batch_size; 

    void* input = readInputBatch(input_path.c_str(),0,start,end,3,32,32); 

    void* var_0 = tensorConvolution(input, conv2d_1_w, 0, 0, 1, 1, 1, 1); 
    void* var_1 = tensorAdd(var_0, conv2d_1_b); 
    void* var_2 = tensorRelu(var_1); 
    void* var_3 = tensorConvolution(var_2, conv2d_2_w, 0, 0, 1, 1, 1, 1); 
    void* var_4 = tensorAdd(var_3, conv2d_2_b); 
    void* var_5 = tensorRelu(var_4); 
    void* var_6 = tensorPooling(var_5,0,2,2,0,0,2,2); 
    void* var_8 = tensorConvolution(var_6, conv2d_3_w, 0, 0, 1, 1, 1, 1); 
    void* var_9 = tensorAdd(var_8, conv2d_3_b); 
    void* var_10 = tensorRelu(var_9); 
    void* var_11 = tensorConvolution(var_10, conv2d_4_w, 0, 0, 1, 1, 1, 1); 
    void* var_12 = tensorAdd(var_11, conv2d_4_b); 
    void* var_13 = tensorRelu(var_12); 
    void* var_14 = tensorPooling(var_13,0,2,2,0,0,2,2); 
    void* var_17 = tensorGemmGPU(var_14, dense_1_w); 
    void* var_18 = tensorAdd(var_17, dense_1_b); 
    void* var_19 = tensorRelu(var_18); 
    void* var_21 = tensorGemmGPU(var_19, dense_2_w); 
    void* var_22 = tensorAdd(var_21, dense_2_b); 
    void* var_23 = tensorSoftmax(var_22); 

    uint32_t* labels = readLabelsBatch3(labels_path.c_str(),start,end); 

    float accuracy = computeAccuracy3(labels, var_23); 
    final_accuracy += accuracy; 
    freeBatchMemory(); 
 
  }

  final_accuracy = final_accuracy / batch_count; 
  dumpFinalAccuracy(final_accuracy); 


  llvm_hpvm_cleanupTensorRt(); 

  return 0; 

}
